"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const tokenizer_1 = require("../tokenizer");
const types_1 = require("../tokenizer/types");
const whitespace_1 = require("../util/whitespace");
// ## Parser utilities
class UtilParser extends tokenizer_1.default {
    // Tests whether parsed token is a contextual keyword.
    isContextual(name) {
        return this.match(types_1.types.name) && this.state.value === name;
    }
    isLookaheadContextual(name) {
        const l = this.lookaheadTypeAndValue();
        return l.type === types_1.types.name && l.value === name;
    }
    // Consumes contextual keyword if possible.
    eatContextual(name) {
        return this.state.value === name && this.eat(types_1.types.name);
    }
    // Asserts that following token is given contextual keyword.
    expectContextual(name, message) {
        if (!this.eatContextual(name))
            this.unexpected(null, message);
    }
    // Test whether a semicolon can be inserted at the current position.
    canInsertSemicolon() {
        return this.match(types_1.types.eof) || this.match(types_1.types.braceR) || this.hasPrecedingLineBreak();
    }
    hasPrecedingLineBreak() {
        const prevToken = this.state.tokens[this.state.tokens.length - 1];
        const lastTokEnd = prevToken ? prevToken.end : 0;
        return whitespace_1.lineBreak.test(this.input.slice(lastTokEnd, this.state.start));
    }
    isLineTerminator() {
        return this.eat(types_1.types.semi) || this.canInsertSemicolon();
    }
    // Consume a semicolon, or, failing that, see if we are allowed to
    // pretend that there is a semicolon at this position.
    semicolon() {
        if (!this.isLineTerminator())
            this.unexpected(null, types_1.types.semi);
    }
    // Expect a token of a given type. If found, consume it, otherwise,
    // raise an unexpected token error at given pos.
    expect(type, pos) {
        const matched = this.eat(type);
        if (!matched) {
            this.unexpected(pos, type);
        }
    }
    // Raise an unexpected token error. Can take the expected token type
    // instead of a message string.
    unexpected(pos = null, messageOrType = "Unexpected token") {
        if (typeof messageOrType !== "string") {
            messageOrType = `Unexpected token, expected "${messageOrType.label}"`;
        }
        throw this.raise(pos != null ? pos : this.state.start, messageOrType);
    }
}
exports.default = UtilParser;
